#ifndef __aspect__lev_h
#define __aspect__lev_h

#include <aspect/material_model/interface.h>
#include <aspect/material_model/simple.h>
#include <aspect/postprocess/interface.h>
#include <aspect/termination_criteria/interface.h>
#include <aspect/simulator_access.h>
#include <aspect/simulator.h>
#include <aspect/global.h>

#include <deal.II/base/quadrature_lib.h>
#include <deal.II/fe/fe_values.h>
#include <deal.II/dofs/dof_tools.h>

#include <aspect/particle/generator/probability_density_function.h>
#include <deal.II/base/geometry_info.h>

#include <boost/lexical_cast.hpp>

namespace aspect
{
  namespace MaterialModel
  {
    using namespace dealii;

    template <int dim>
    class Lev : public MaterialModel::Simple<dim>
    {
      public:
        virtual void evaluate (const MaterialModel::MaterialModelInputs<dim> &in,
                               MaterialModel::MaterialModelOutputs<dim> &out) const;
        static void declare_parameters (ParameterHandler &prm);
        virtual void parse_parameters (ParameterHandler &prm);
        virtual bool is_compressible () const;
        virtual double reference_viscosity () const;
        virtual double reference_density () const;
      private:
        double eta_N, eta_0;
        static double delta (const unsigned int i, const unsigned int j);
    };
  }

  namespace Particle
  {
    namespace Generator
    {
      template <int dim>
      class EvaluationPoints : public Interface<dim>
      {
        public:
          virtual
          void
          generate_particles(std::multimap<types::LevelInd, Particle<dim> > &particles);

          virtual
          void
          parse_parameters (ParameterHandler &prm);

        private:
          std::vector<Point<dim> > evaluation_points;
      };

    }
  }

  namespace Postprocess
  {
      template <int dim>
      class InterfaceDepth : public Interface<dim>, public ::aspect::SimulatorAccess<dim>
      {
          public:
              virtual
              std::pair<std::string, std::string>
              execute(TableHandler &statistics);
      };
  }

  namespace TerminationCriteria
  {
      template <int dim>
      class InterfaceDepth : public Interface<dim>, public SimulatorAccess<dim>
      {
          public:
            static void declare_parameters (ParameterHandler &prm);
            virtual void parse_parameters (ParameterHandler &prm);
            virtual bool execute(void);
          private:
            double terminate_depth;
            std::string composition_name;
      };
  }
}

#endif

namespace aspect
{
  namespace MaterialModel
  {
    template <int dim>
    void
    Lev<dim>::evaluate (const MaterialModel::MaterialModelInputs<dim> &in,
                        MaterialModel::MaterialModelOutputs<dim> &out) const
    {
      Simple<dim>::evaluate(in,out);

      AssertThrow((this->introspection().compositional_name_exists("gamma")),
                  ExcMessage("Lev and Hager model only works if there is a compositional field called gamma."));
      AssertThrow(this->introspection().compositional_name_exists("ni"),
                  ExcMessage("Lev and Hager model only works if there is a compositional field called ni."));
      AssertThrow(this->introspection().compositional_name_exists("nj"),
                  ExcMessage("Lev and Hager model only works if there is a compositional field called nj."));
      if (dim == 3)
        AssertThrow(this->introspection().compositional_name_exists("nk"),
                    ExcMessage("Lev and Hager model only works if there is a compositional field called nk."));

      const unsigned int c_idx_gamma = this->introspection().compositional_index_for_name("gamma");

      std::vector<unsigned int> c_idx_n;
      c_idx_n.push_back (this->introspection().compositional_index_for_name("ni"));
      c_idx_n.push_back (this->introspection().compositional_index_for_name("nj"));
      if (dim == 3)
        c_idx_n.push_back (this->introspection().compositional_index_for_name("nk"));

      // Get the grad_u tensor, at the center of this cell, if possible.
      Tensor<2,dim> grad_u;
      if (in.current_cell.state() == IteratorState::valid)
        {
          const QMidpoint<dim> quadrature_formula;
          FEValues<dim> fe_values (this->get_mapping(),
                                   this->get_fe(),
                                   quadrature_formula,
                                   update_gradients);
          fe_values.reinit(in.current_cell);
          std::vector<Tensor<2,dim> > velocity_gradients (quadrature_formula.size());
          fe_values[this->introspection().extractors.velocities]
          .get_function_gradients(this->get_solution(), velocity_gradients);

          grad_u = velocity_gradients[0];
        }

      for (unsigned int q=0; q<in.position.size(); ++q)
        {
          out.densities[q] = (in.composition[q][c_idx_gamma] > 0.1 ? 1 : 0);
          out.viscosities[q] = eta_N;
          out.thermal_expansion_coefficients[q] = 0;
          out.specific_heat[q] = 0;
          out.thermal_conductivities[q] = 0;
          out.compressibilities[q] = 0.0;
          out.entropy_derivative_pressure[q] = 0.0;
          out.entropy_derivative_temperature[q] = 0.0;
          for (unsigned int c=0; c<in.composition[q].size(); ++c)
              out.reaction_terms[q][c] = 0.0;

          Point<dim> n;
          for (unsigned int i=0; i<dim; ++i)
            n[i] = in.composition[q][c_idx_n[i]];

          // Symmetric and anti-symmetric parts of grad_u
          const SymmetricTensor<2,dim> D = symmetrize(grad_u);
          Tensor<2,dim> W;
          for (unsigned int i=0; i<dim; ++i)
            for (unsigned int j=0; j<dim; ++j)
              W[i][j] = grad_u[i][j] - D[i][j];

          Point<dim> n_dot;
          for (unsigned int i=0; i<dim; ++i)
            {
              n_dot[i] = 0.0;
              // outer summation over each value of j.
              for (unsigned int j=0; j<dim; ++j)
              {
                float Wn = W[i][j];
                for (unsigned int k=0; k<dim; ++k)
                  {
                    Wn -= D[k][i]*n[k]*n[j] - D[k][j]*n[k]*n[i];
                  }
                n_dot[i] += Wn * n[j];
              }
            }

          for (unsigned int i=0; i<dim; ++i)
            out.reaction_terms[q][c_idx_n[i]] = n_dot[i] * this->get_timestep();

          if ((n[0] != 0) || (n[1] != 0) || (dim == 3 && n[2] != 0))
            {
              SymmetricTensor<4,dim> Lambda;
              for (unsigned int i=0; i<dim; ++i)
                for (unsigned int j=0; j<dim; ++j)
                  for (unsigned int k=0; k<dim; ++k)
                    for (unsigned int l=0; l<dim; ++l)
                      Lambda[i][j][k][l] = 1./2. * (n[i]*n[k]*delta(l,j)
                                                    + n[j]*n[k]*delta(i,l)
                                                    + n[i]*n[l]*delta(k,j)
                                                    + n[j]*n[l]*delta(i,k))
                                           - 2*n[i]*n[j]*n[k]*n[l];
              out.stress_strain_directors[q] = dealii::identity_tensor<dim> ()
                                               - (1. - eta_0) * Lambda;
            }
        }
    }

    template <int dim>
    double
    Lev<dim>::delta (const unsigned int i,
                     const unsigned int j)
    {
      return (i == j ? 1 : 0);
    }

    template <int dim>
    bool
    Lev<dim>::is_compressible () const
    {
      return false;
    }

    template <int dim>
    double
    Lev<dim>::reference_density () const
    {
      return 1.0;
    }

    template <int dim>
    double
    Lev<dim>::reference_viscosity () const
    {
      return 1.0;
    }

    template <int dim>
    void
    Lev<dim>::parse_parameters (ParameterHandler &prm)
    {
      prm.enter_subsection("Material model");
      {
        prm.enter_subsection("Lev");
        {
          eta_N = prm.get_double("Normal viscosity");
          eta_0 = prm.get_double("Shear viscosity")/eta_N;
        }
        prm.leave_subsection();
      }
      prm.leave_subsection();
    }

    template <int dim>
    void
    Lev<dim>::declare_parameters (ParameterHandler &prm)
    {
      prm.enter_subsection("Material model");
      {
        prm.enter_subsection("Lev");
        {
          prm.declare_entry ("Normal viscosity", "1e-3",
                             Patterns::Double(),
                             "Magnitude of normal viscosity.");
          prm.declare_entry ("Shear viscosity", "1e-4",
                             Patterns::Double(),
                             "Magnitude of shear viscosity.");
        }
        prm.leave_subsection();
      }
      prm.leave_subsection();
    }
  }

  namespace Particle
  {
    namespace Generator
    {
      template <int dim>
      void
      EvaluationPoints<dim>::generate_particles(std::multimap<types::LevelInd, Particle<dim> > &particles)
      {
        types::particle_index particle_index = 0;

        for (unsigned int i=0; i < evaluation_points.size(); ++i)
          {
            try
              {
                particles.insert(this->generate_particle(evaluation_points[i],particle_index));
              }
            catch (ExcParticlePointNotInDomain &)
              {}

            particle_index++;
          }
      }

      template <int dim>
      void
      EvaluationPoints<dim>::parse_parameters (ParameterHandler &prm)
      {
        prm.enter_subsection("Postprocess");
        {
          prm.enter_subsection("Point values");
          {
            const std::vector<std::string> point_list
              = Utilities::split_string_list(prm.get("Evaluation points"), ';');
            for (unsigned int p=0; p<point_list.size(); ++p)
              {
                const std::vector<std::string> coordinates
                  = Utilities::split_string_list(point_list[p], ',');
                AssertThrow (coordinates.size() == dim,
                             ExcMessage ("In setting up the list of evaluation points for the <Point values> "
                                         "postprocessor, one of the evaluation points reads <"
                                         + point_list[p] +
                                         ">, but this does not correspond to a list of numbers with "
                                         "as many coordinates as you run your simulation in."));
  
                Point<dim> point;
                for (unsigned int d=0; d<dim; ++d)
                  point[d] = Utilities::string_to_double (coordinates[d]);
                evaluation_points.push_back (point);
              }
          }
          prm.leave_subsection();
        }
        prm.leave_subsection();
      }
    }
  }

  namespace Postprocess
  {
      template <int dim>
      std::pair<std::string,std::string>
      InterfaceDepth<dim>::execute (TableHandler &statistics)
      {
          const QGauss<dim> quadrature_formula (this->get_fe()
                                            .base_element(this->introspection().base_elements.compositional_fields).degree+1);
          const unsigned int n_q_points = quadrature_formula.size();

          FEValues<dim> fe_values (this->get_mapping(),
                                   this->get_fe(),
                                   quadrature_formula,
                                   update_values   |
                                   update_quadrature_points |
                                   update_JxW_values);
          std::vector<double> gamma_values(n_q_points);
          std::vector<Tensor<1,dim> > velocity_values(n_q_points);

          double local_max_depth = 0;
          double local_max_velocity = 0;
          const unsigned int c_idx_gamma = this->introspection().compositional_index_for_name("gamma");

          typename DoFHandler<dim>::active_cell_iterator
          cell = this->get_dof_handler().begin_active(),
          endc = this->get_dof_handler().end();
          for (; cell!=endc; ++cell)
            if (cell->is_locally_owned())
              {
                fe_values.reinit (cell);
                fe_values[this->introspection().extractors.compositional_fields[c_idx_gamma]].get_function_values (this->get_solution(), gamma_values);
                fe_values[this->introspection().extractors.velocities].get_function_values (this->get_solution(), velocity_values);

                for (unsigned int q = 0; q < n_q_points; ++q)
                  {
                    const double depth = this->get_geometry_model().depth(fe_values.quadrature_point(q));
                    if (gamma_values[q] > 0.5 && depth > local_max_depth) {
                        local_max_depth = depth;
                        //local_max_velocity = std::sqrt(velocity_values[q]*velocity_values[q]);
                        local_max_velocity = std::sqrt(velocity_values[q][dim-1]*velocity_values[q][dim-1]);
                    }
                  }
              }

          const double global_max_depth
              = Utilities::MPI::max(local_max_depth, this->get_mpi_communicator());

          // Find interface velocity from MPI process which contains the deepest part of the interface
          const double known_local_max_velocity
              = (local_max_depth > global_max_depth-1e-10) ? local_max_velocity : 0.0;
          const double global_max_velocity
              = Utilities::MPI::max(known_local_max_velocity, this->get_mpi_communicator());

          statistics.add_value("Maximum interface depth", global_max_depth);
          statistics.add_value("Maximum interface velocity", global_max_velocity);

          std::ostringstream output;
          output.precision(3);
          output << global_max_depth << " m, "
                 << global_max_velocity << " m/s";

          return std::pair<std::string, std::string> ("Maximum interface depth/velocity", output.str());
      }
  }

  namespace TerminationCriteria
  {
      template <int dim>
      bool
      InterfaceDepth<dim>::execute()
      {
          const QGauss<dim> quadrature_formula (this->get_fe()
                                            .base_element(this->introspection().base_elements.compositional_fields).degree+1);
          const unsigned int n_q_points = quadrature_formula.size();

          FEValues<dim> fe_values (this->get_mapping(),
                                   this->get_fe(),
                                   quadrature_formula,
                                   update_values   |
                                   update_quadrature_points |
                                   update_JxW_values);
          std::vector<double> gamma_values(n_q_points);

          double local_max_depth = 0;
          const unsigned int c_idx_gamma = this->introspection().compositional_index_for_name(composition_name);

          typename DoFHandler<dim>::active_cell_iterator
          cell = this->get_dof_handler().begin_active(),
          endc = this->get_dof_handler().end();
          for (; cell!=endc; ++cell)
            if (cell->is_locally_owned())
              {
                fe_values.reinit (cell);
                fe_values[this->introspection().extractors.compositional_fields[c_idx_gamma]].get_function_values (this->get_solution(),
                                                                                                                   gamma_values);
                for (unsigned int q = 0; q < n_q_points; ++q)
                  {
                    const double depth = this->get_geometry_model().depth(fe_values.quadrature_point(q));
                    if (gamma_values[q] > 0.5 && depth > local_max_depth)
                        local_max_depth = depth;
                  }
              }

          const double global_max_depth
              = Utilities::MPI::max(local_max_depth, this->get_mpi_communicator());          

          return (global_max_depth > terminate_depth);
      }

      template <int dim>
      void
      InterfaceDepth<dim>::declare_parameters (ParameterHandler &prm)
      {
          prm.enter_subsection("Termination criteria");
          {
              prm.enter_subsection("Interface depth");
              {
                  prm.declare_entry("Termination depth", "0.5",
                                    Patterns::Double (),
                                    "Terminate model when composition passes this depth.");
                  prm.declare_entry("Composition name", "",
                                    Patterns::Anything (),
                                    "Use this compositional field.");
              }
              prm.leave_subsection();
          }
          prm.leave_subsection();
      }

      template <int dim>
      void
      InterfaceDepth<dim>::parse_parameters (ParameterHandler &prm)
      {
          prm.enter_subsection("Termination criteria");
          {
              prm.enter_subsection("Interface depth");
              {
                  terminate_depth = prm.get_double("Termination depth");
                  composition_name = prm.get("Composition name");
              }
              prm.leave_subsection();
          }
          prm.leave_subsection();
      }
  }
}

namespace aspect
{
  namespace MaterialModel
  {
    ASPECT_REGISTER_MATERIAL_MODEL(Lev,
                                   "Lev and Hager",
                                   "Einat Lev model (nondimensionalized)")
  }

  namespace Particle
  {
    namespace Generator
    {
      ASPECT_REGISTER_PARTICLE_GENERATOR(EvaluationPoints,
                                         "evaluation points",
                                         "Generates particles at same points that the "
                                         "'point values' postprocessor reads.")
    }
  }

  namespace Postprocess
  {
      ASPECT_REGISTER_POSTPROCESSOR(InterfaceDepth,
                                    "interface depth",
                                    "A postprocessor that outputs the maximum "
                                    "depth of the 'gamma' compositional field.")
  }

  namespace TerminationCriteria
  {
    ASPECT_REGISTER_TERMINATION_CRITERION(InterfaceDepth,
                                         "interface depth",
                                         "Termination when compositional field "
                                         "advects past specified depth")
  }
}
