#include <aspect/material_model/interface.h>
#include <aspect/material_model/multicomponent.h>
#include <aspect/simulator_access.h>
#include <aspect/simulator.h>
#include <aspect/global.h>
#include <aspect/simulator_signals.h>
#include <aspect/gravity_model/interface.h>
#include <aspect/postprocess/visualization.h>

#include <deal.II/dofs/dof_tools.h>
#include <deal.II/numerics/data_out.h>
#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/function_lib.h>
#include <deal.II/numerics/error_estimator.h>
#include <deal.II/numerics/vector_tools.h>
#include <deal.II/numerics/data_postprocessor.h>
#include <deal.II/fe/fe_values.h>

#include <deal.II/base/parsed_function.h>

/**
 * This file contains several things.
 *
 * 1. A GravityModel to prescribe RHS of stokes equation
 *
 * 2. A DataPostprocessorVector to visualize the difference
 *    between the computed solution, and the expected
 *    analytic solution.
 *
 * 3. A regular Postprocessor to print the L1 and L2 error
 *    norms to the statistics file.
 *
 * 4. A Signal hack to disable the pressure solution variable.
 *    This forces our prescribed RHS to be accomodated in the
 *    strain rate, and not in the pressure, which makes it easier
 *    to test a constitutive relationship.
 */

namespace aspect
{

  // Prescribe RHS.
  //
  // Assumes that density is set to 1 everywhere.
  // Used to prescribe a known RHS to Stokes flow (effectively
  // sets grad(P) in a constant-density Stokes flow,
  // so long as the actual 'P' solution variable is
  // disabled via signaling hack)
  namespace GravityModel
  {
    using namespace dealii;

    template <int dim>
    class ManufacturedRHS : public Interface<dim>
    {
      public:
        ManufacturedRHS ();

        virtual
        Tensor<1,dim> gravity_vector (const Point<dim> &position) const;

        virtual void parse_parameters (ParameterHandler &prm);
      private:
        double eta_S;
        double eta_N;
        std_cxx11::unique_ptr<Functions::ParsedFunction<dim> > function;
    };

    template <int dim>
    ManufacturedRHS<dim>::ManufacturedRHS ()
    {}

    template <int dim>
    Tensor<1, dim>
    ManufacturedRHS<dim>::
    gravity_vector (const Point<dim> &position) const
    {
      Tensor<1, dim> gravity;
      const double x = position(0);
      const double y = position(1);

      const double u = (cos(2*x)+1)*cos(2*y);
      const double dudx = -2*sin(2*x)*cos(2*y);
      const double dudy = -2*(cos(2*x)+1)*sin(2*y);
      const double d2udx2 = -4*cos(2*x)*cos(2*y);
      const double dudxdy = 4*sin(2*x)*sin(2*y);
      const double dudydx = 4*sin(2*x)*sin(2*y);
      const double d2udy2 = -4*(cos(2*x)+1)*cos(2*y);

      const double v = sin(2*x)*sin(2*y);
      const double dvdx = 2*cos(2*x)*sin(2*y);
      const double dvdy = 2*sin(2*x)*cos(2*y);
      const double d2vdx2 = -4*sin(2*x)*sin(2*y);
      const double dvdxdy = 4*cos(2*x)*cos(2*y);
      const double dvdydx = 4*cos(2*x)*cos(2*y);
      const double d2vdy2 = -4*sin(2*x)*sin(2*y);

      // eta = | A  B |
      //       | C  D |
      const double a = dudx;
      const double dadx = d2udx2;
      const double dady = dudydx;
      const double b = dvdx;
      const double dbdx = d2vdx2;
      const double dbdy = dvdydx;
      const double c = dudy;
      const double dcdx = dudxdy;
      const double dcdy = d2udy2;
      const double d = dvdy;
      const double dddx = dvdxdy;
      const double dddy = d2vdy2;

      // RHS
      //
      // C*eta = | L0000A+L0001B+L0010C+L0011D  L0100A+L0101B+L0110C+L0111D |
      //         |                                                          |
      //         | L1000A+L1001B+L1010C+L1011D  L1100A+L1101B+L1110C+L1111D |

      const double n0 = function->value(position,0);
      const double n1 = function->value(position,1);

      const double C0000 = 1.0 - (1 - eta_S/eta_N)*(0.5*(n0*n0+n0*n0+n0*n0+n0*n0) - 2*(n0*n0*n0*n0));
      const double C0001 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n0*n1+n0*n1) - 2*(n0*n0*n0*n1));
      const double C0010 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n0*n1+n0*n1) - 2*(n0*n0*n1*n0));
      const double C0011 = 0.0 - (1 - eta_S/eta_N)*(0.5*(0) - 2*(n0*n0*n1*n1));
      const double C0100 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n1*n0+n1*n0) - 2*(n0*n1*n0*n0));
      const double C0101 = 0.5 - (1 - eta_S/eta_N)*(0.5*(n0*n0+n1*n1) - 2*(n0*n1*n0*n1));
      const double C0110 = 0.5 - (1 - eta_S/eta_N)*(0.5*(n1*n1+n0*n0) - 2*(n0*n1*n1*n0));
      const double C0111 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n0*n1+n0*n1) - 2*(n0*n1*n1*n1));
      const double C1000 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n1*n0+n1*n0) - 2*(n1*n0*n0*n0));
      const double C1001 = 0.5 - (1 - eta_S/eta_N)*(0.5*(n0*n0+n1*n1) - 2*(n1*n0*n0*n1));
      const double C1010 = 0.5 - (1 - eta_S/eta_N)*(0.5*(n1*n1+n0*n0) - 2*(n1*n0*n1*n0));
      const double C1011 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n0*n1+n0*n1) - 2*(n1*n0*n1*n1));
      const double C1100 = 0.0 - (1 - eta_S/eta_N)*(0.5*(0) - 2*(n1*n1*n0*n0));
      const double C1101 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n1*n0+n1*n0) - 2*(n1*n1*n0*n1));
      const double C1110 = 0.0 - (1 - eta_S/eta_N)*(0.5*(n1*n0+n1*n0) - 2*(n1*n1*n1*n0));
      const double C1111 = 1.0 - (1 - eta_S/eta_N)*(0.5*(n1*n1+n1*n1+n1*n1+n1*n1) - 2*(n1*n1*n1*n1));

      gravity[0] = C0000*dadx+C0001*dbdx+C0010*dcdx+C0011*dddx +
                   C1000*dady+C1001*dbdy+C1010*dcdy+C1011*dddy;
      gravity[1] = C0100*dadx+C0101*dbdx+C0110*dcdx+C0111*dddx +
                   C1100*dady+C1101*dbdy+C1110*dcdy+C1111*dddy;

      return -2*gravity;
    }

    template <int dim>
    void
    ManufacturedRHS<dim>::parse_parameters (ParameterHandler &prm)
    {
      prm.enter_subsection("Material model");
      {
        prm.enter_subsection("Lev");
        {
          eta_N = prm.get_double("Normal viscosity");
          eta_S = prm.get_double("Shear viscosity");
        }
        prm.leave_subsection();
      }
      prm.leave_subsection();

      prm.enter_subsection("Initial composition model");
      {
        prm.enter_subsection("Function");
        try
          {
            function.reset (new Functions::ParsedFunction<dim>(2));
            function->parse_parameters (prm);
          }
        catch (...)
          {
            std::cerr << "ERROR: FunctionParser failed to parse\n"
                      << "\t'Initial composition model.Function'\n"
                      << "with expression\n"
                      << "\t'" << prm.get("Function expression") << "'\n"
                      << "More information about the cause of the parse error \n"
                      << "is shown below.\n";
            throw;
          }

        prm.leave_subsection();
      }
      prm.leave_subsection();
    }
  }

  // Postprocessor to compare numerical solution
  // with analytic solution
  namespace Postprocess
  {
    using namespace dealii;

    // A postprocessor to visualize the difference between
    // the computed solution and the known analytic solution.
    namespace VisualizationPostprocessors
    {

      template <int dim>
      class ErrorVis
        : public DataPostprocessorVector<dim>,
          public SimulatorAccess<dim>,
          public Interface<dim>
      {
        public:
          ErrorVis ();

          virtual
          void
          compute_derived_quantities_vector (const std::vector<Vector<double> >              &uh,
                                             const std::vector<std::vector<Tensor<1,dim> > > &duh,
                                             const std::vector<std::vector<Tensor<2,dim> > > &dduh,
                                             const std::vector<Point<dim> >                  &normals,
                                             const std::vector<Point<dim> >                  &evaluation_points,
                                             std::vector<Vector<double> >                    &computed_quantities) const;
      };

      template <int dim>
      ErrorVis<dim>::
      ErrorVis ()
        :
        DataPostprocessorVector<dim> ("error",
                                      update_values | update_q_points)
      {}

      template <int dim>
      void
      ErrorVis<dim>::
      compute_derived_quantities_vector (const std::vector<Vector<double> >              &uh,
                                         const std::vector<std::vector<Tensor<1,dim> > > &,
                                         const std::vector<std::vector<Tensor<2,dim> > > &,
                                         const std::vector<Point<dim> > &,
                                         const std::vector<Point<dim> >                  &evaluation_points,
                                         std::vector<Vector<double> >                    &computed_quantities) const
      {
        const unsigned int n_quadrature_points = uh.size();
        Assert (computed_quantities.size() == n_quadrature_points,    ExcInternalError());
        Assert (computed_quantities[0].size() == dim,                   ExcInternalError());
        Assert (uh[0].size() == this->introspection().n_components,           ExcInternalError());

        for (unsigned int q=0; q<n_quadrature_points; ++q)
          {
            const double x = evaluation_points[q](0);
            const double y = evaluation_points[q](1);

            // ANALYTIC SOLUTION
            const double analytic_vx = (std::cos(2*x)+1)*std::cos(2*y);
            const double analytic_vy = std::sin(2*x)*std::sin(2*y);

            const double numerical_vx = uh[q][this->introspection().component_indices.velocities[0]];
            const double numerical_vy = uh[q][this->introspection().component_indices.velocities[1]];

            computed_quantities[q](0) = numerical_vx - analytic_vx;
            computed_quantities[q](1) = numerical_vy - analytic_vy;
          }
      }
    }

    // We're still in the Postprocess namespace.
    // Now we want a postprocessor to give us some quantitative
    // measure of how bad our solution is. We return the L1
    // and L2 norms of our computed solution compared with
    // the analytic solution.
    namespace AnalyticSolutions
    {
      template <int dim>
      class ManufacturedAnalyticSol : public Function<dim>
      {
        public:
          ManufacturedAnalyticSol ()
            :
            Function<dim>()
          {}

          virtual void
          vector_value (const Point<dim> &position,
                        Vector<double> &values) const
          {
            AssertDimension(values.size(), 4);
            const double x = position(0);
            const double y = position(1);

            // ANALYTIC SOLUTION
            values[0] = (std::cos(2*x)+1)*std::cos(2*y);
            values[1] = std::sin(2*x)*std::sin(2*y);
          }
      };
    }

    template <int dim>
    class NormPostprocessor
      : public Interface<dim>,
        public SimulatorAccess<dim>
    {
      public:
        virtual
        std::pair<std::string,std::string>
        execute (TableHandler &statistics);
    };

    template <int dim>
    std::pair<std::string,std::string>
    NormPostprocessor<dim>::execute (TableHandler &statistics)
    {
      AssertThrow(Utilities::MPI::n_mpi_processes(this->get_mpi_communicator()) == 1,
                  ExcNotImplemented());

      std_cxx1x::shared_ptr<Function<dim> > ref_func;

      ref_func.reset (new AnalyticSolutions::ManufacturedAnalyticSol<dim>());

      const QGauss<dim> quadrature_formula
      (this->get_fe().base_element(this->introspection().base_elements.velocities).degree+2);

      Vector<float> cellwise_errors_u (this->get_triangulation().n_active_cells());
      Vector<float> cellwise_errors_ul2 (this->get_triangulation().n_active_cells());

      ComponentSelectFunction<dim> comp_u(std::pair<unsigned int, unsigned int>(0,dim),
                                          dim+2);

      VectorTools::integrate_difference (this->get_mapping(),
                                         this->get_dof_handler(),
                                         this->get_solution(),
                                         *ref_func,
                                         cellwise_errors_u,
                                         quadrature_formula,
                                         VectorTools::L1_norm,
                                         &comp_u);
      VectorTools::integrate_difference (this->get_mapping(),
                                         this->get_dof_handler(),
                                         this->get_solution(),
                                         *ref_func,
                                         cellwise_errors_ul2,
                                         quadrature_formula,
                                         VectorTools::L2_norm,
                                         &comp_u);

      statistics.add_value ("Error L1 Norm", cellwise_errors_u.l1_norm());
      statistics.set_precision ("Error L1 Norm", 8);
      statistics.set_scientific ("Error L1 Norm", true);

      statistics.add_value ("Error L2 Norm", cellwise_errors_ul2.l2_norm());
      statistics.set_precision ("Error L2 Norm", 8);
      statistics.set_scientific ("Error L2 Norm", true);

      std::ostringstream os;
      os << std::scientific << cellwise_errors_u.l1_norm()
         << ", " << cellwise_errors_ul2.l2_norm();

      return std::make_pair("Errors u_L1, u_L2:", os.str());
    }
  }

  // We want to test the constitutive law only, so we don't want any
  // of our prescribed RHS to be accomodated by the pressure field.
  // To this end we want to effectively disable that solution variable.
  // (Note, this is similar to solving for a constant density Stokes
  // flow, where dynamic pressures are the only driver of flow. In this case
  // we can think of the gravity field as grad(P), and ignore aspect's P
  // solution variable)
  //
  // Prescribe zero pressure everywhere. This function is
  // called by a signal while the assembly function is applying
  // constraints to the stiffness matrix. It adds constraints to
  // the Pressure solution everywhere.
  template <int dim>
  void prescribe_zero_pressures (const SimulatorAccess<dim> &simulator_access,
                                 ConstraintMatrix &current_constraints)
  {
    Quadrature<dim> quadrature (simulator_access.get_fe().get_unit_support_points());
    FEValues<dim> fe_values (simulator_access.get_fe(), quadrature, update_q_points);
    typename DoFHandler<dim>::active_cell_iterator cell;

    for (cell = simulator_access.get_dof_handler().begin_active();
         cell != simulator_access.get_dof_handler().end();
         ++cell)
      if (! cell->is_artificial())
        {
          fe_values.reinit(cell);
          std::vector<unsigned int> local_dof_indices(simulator_access.get_fe().dofs_per_cell);
          cell->get_dof_indices (local_dof_indices);

          for (unsigned int q=0; q<quadrature.size(); ++q)
            if (current_constraints.can_store_line(local_dof_indices[q]) &&
                !current_constraints.is_constrained(local_dof_indices[q]))
              {
                const unsigned int c_idx =
                  simulator_access.get_fe().system_to_component_index(q).first;

                if (c_idx == simulator_access.introspection().component_indices.pressure)
                  {
                    current_constraints.add_line (local_dof_indices[q]);
                    current_constraints.set_inhomogeneity (local_dof_indices[q], 0.0);
                  }
              }
        }
  }

  template <int dim>
  void signal_connector (SimulatorSignals<dim> &signals)
  {
    signals.post_constraints_creation.connect (&prescribe_zero_pressures<dim>);
  }
}


namespace aspect
{
  namespace Postprocess
  {
    ASPECT_REGISTER_POSTPROCESSOR(NormPostprocessor,
                                  "convergence",
                                  "Test L1 and L2 norms between computed "
                                  "solution and analytic solution.")

    namespace VisualizationPostprocessors
    {
      ASPECT_REGISTER_VISUALIZATION_POSTPROCESSOR(ErrorVis,
                                                  "error",
                                                  "Visualize difference between computed "
                                                  "solution and analytic solution")
    }
  }

  namespace GravityModel
  {
    ASPECT_REGISTER_GRAVITY_MODEL(ManufacturedRHS,
                                  "manufactured",
                                  "Manufactured solution of grad^2(u) for RHS.")
  }

  ASPECT_REGISTER_SIGNALS_CONNECTOR(signal_connector<2>, signal_connector<3>)
}
