#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from os.path import expanduser
sys.path.insert(0, expanduser("~")+'/.local/lib/python2.7/site-packages')

import pylab as pl
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.colors import LogNorm

h1, n1 = np.mgrid[0:1:500j,-15:15:1000j]
n1 = 10**(n1)

extents=(np.log10(n1).min(),np.log10(n1).max(),h1.min(),h1.max())

h2=1.0
n2=1.0
nn = (n1*h1+n2*h2)/(h1+h2)
ns = (h1+h2)/(h1/n1 + h2/n2)
div = max(abs(np.log10(np.array([nn.min(), ns.min(), ns.max(), nn.max()]))))
viscvmin = 10**(-1*div)
viscvmax = 10**(div)

manual_locations = zip([-8,-4,4,8], np.ones(4)*0.5)

def plot_schematic(ax):
    try:
        schematic = pl.imread('simplified_schematic.png')
    except IOError:
        schematic = np.zeros((682,354,4))
    w = np.where(schematic[:,:,3]==0.0)
    for i in range(4):
        schematic[:,:,i][w] = 1.0
    aspect = float(schematic.shape[0])/schematic.shape[1]
    aspect *= (extents[1]-extents[0])/(extents[3]-extents[2])
    img = ax.imshow(schematic,extent=extents, aspect=aspect)
    ax.set_axis_off()
    pl.setp(ax.get_xticklabels(), visible=False)
    pl.setp(ax.get_yticklabels(), visible=False)
    return img

def plot_av_magnitude(ax,cmap):
    Z = ns/nn
    img = ax.contourf(Z, extent=extents, origin='lower', cmap=cmap, norm=LogNorm(vmin=Z.min(), vmax=Z.max()))
    CS = ax.contour(np.log10(n1), h1, (Z), colors='k', norm=LogNorm(vmin=Z.min(), vmax=Z.max()), linewidths=1)
    xs = [-14, -8, -2, 2, 8, 14]
    manual_locations = zip(xs, np.ones(len(xs))*0.5)
    pl.clabel(CS, inline=1, fmt='%.0e', manual=manual_locations)
    ax.set_title('AV magnitude $(\\eta_S/\\eta_N)$')
    return img

def plot_shear_visc(ax,cmap):
    foobar = ns*0+viscvmin
    foobar[0][1] = viscvmax
    img = ax.contourf(foobar, extent=extents, origin='lower', norm=LogNorm(vmin=viscvmin, vmax=viscvmax), cmap=cmap)
    ax.contourf(ns, extent=extents, origin='lower', norm=LogNorm(vmin=viscvmin, vmax=viscvmax), cmap=cmap)
    CS = ax.contour(np.log10(n1), h1, ns, colors='k', norm=LogNorm(vmin=viscvmin, vmax=viscvmax), levels=(1e-13, 1e-11, 1e-9, 1e-7, 1e-5, 1e-3, 0.1), linewidths=1)
    xs = [-13, -9, -5, -1]
    manual_locations = zip(xs, np.ones(len(xs))*0.5)
    pl.clabel(CS, inline=1, fmt='%0.0e', manual=manual_locations)
    ax.set_title('Shear viscosity ($\\overline{ \\eta}_S$)')
    return img

def plot_normal_visc(ax,cmap):
    foobar = ns*0+viscvmin
    foobar[0][1] = viscvmax
    img = ax.contourf(foobar, extent=extents, origin='lower', norm=LogNorm(vmin=viscvmin, vmax=viscvmax), cmap=cmap)
    ax.contourf(np.log10(n1), h1, nn, norm=LogNorm(vmin=viscvmin, vmax=viscvmax), cmap=cmap)
    CS = ax.contour(np.log10(n1), h1, nn, colors='k', norm=LogNorm(vmin=viscvmin, vmax=viscvmax), levels=(10, 1e3, 1e5, 1e7, 1e9, 1e11, 1e13), linewidths=0.7)
    xs = [1, 5, 9, 13]
    manual_locations = zip(xs, np.ones(len(xs))*0.5)
    pl.clabel(CS, inline=1, fmt='%0.0e', manual=manual_locations)
    ax.set_title('Normal viscosity ($\\overline{ \\eta}_N$)')
    return img

if __name__ == '__main__':
    try:
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import cmocean
        cmap1 = cmocean.cm.speed
        cmap2 = cmocean.cm.balance
    except:
        cmap1 = pl.cm.viridis
        cmap2 = pl.cm.coolwarm

    try:
        pl.style.use('paper')
    except: pass
    fig, axes = pl.subplots(2,2,figsize=(3.5,2.75),dpi=300)

    im0 = plot_schematic    (axes[0][0])
    im1 = plot_av_magnitude (axes[0][1],cmap1)
    im2 = plot_shear_visc   (axes[1][0],cmap2)
    im3 = plot_normal_visc  (axes[1][1],cmap2)

    cax1 = make_axes_locatable(axes[0][1]).append_axes("right", size="5%", pad=0.05)
    pl.colorbar(im1, cax=cax1)

    cax3 = make_axes_locatable(axes[1][1]).append_axes("right", size="5%", pad=0.05)
    pl.colorbar(im3, cax=cax3)

    axes[1][0].set_ylabel('$\\overline{ h}_1$')
    axes[1][0].set_yticks([0,1])
    axes[1][1].set_yticks([])
    axes[0][1].set_yticks([])
    axes[1][0].yaxis.set_label_coords(0.,0.5)
    axes[-1][0].set_xlabel('$log_{10}(\\overline{ \\eta}_1)$')
    axes[-1][1].set_xlabel('$log_{10}(\\overline{ \\eta}_1)$')
    axes[-1][0].xaxis.set_label_coords(0.5,-0.15)
    axes[-1][1].xaxis.set_label_coords(0.5,-0.15)

    pl.tight_layout(h_pad=0.05,w_pad=0.05,pad=0.1)

    pl.savefig('fig1.eps')
