#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Generates figures 3 and 4.
Assumes models have already been run.
If not, do `make models` in Ensemble directory.
'''

import sys
from os.path import expanduser
sys.path.insert(0, expanduser("~")+'/.local/lib/python2.7/site-packages')
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pylab as pl
    import numpy as np
    import scipy.interpolate as interp
    from matplotlib.colors import LogNorm
    import scipy
    from tools import load_ensemble

def discretize_cmap(cmap,N):
    # Generally useful function for converting a colormap into discrete segments.
    colors_i = np.concatenate((np.linspace(0, 1., N), (0.,0.,0.,0.)))
    colors_rgba = cmap(colors_i)
    indices = np.linspace(0, 1., N+1)
    cdict = {}
    for ki,key in enumerate(('red','green','blue')):
        cdict[key] = [ (indices[i], colors_rgba[i-1,ki], colors_rgba[i,ki]) for i in xrange(N+1) ]
    # Return colormap object.
    return pl.matplotlib.colors.LinearSegmentedColormap(cmap.name + "_%d"%N, cdict, 1024)

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import cmocean
    cmap1d = discretize_cmap(cmocean.cm.haline, 23)
    cmap2 = cmocean.cm.delta
    cmap2d = discretize_cmap(cmap2, 15)
except:
    cmap1d = discretize_cmap(pl.cm.viridis, 23)
    cmap2 = pl.cm.coolwarm
    cmap2d = discretize_cmap(cmap2, 15)

class Surface():
    ''' Class for parabolic surface fitting '''
    def __init__(self):
        self.paramguess = np.zeros(9)
        self.data = (None,None)
        self.fit = None

    @staticmethod
    def coeff(params,H,D):
        a,b,c,d,e,f,g,h,i = params
        A = a*H + b*D + c
        B = d*H + e*D + f
        C = g*H + h*D + i
        return A,B,C

    @staticmethod
    def F(params,k,h,d):
        A,B,C = Surface.coeff(params,h,d)
        return A*k**2 + B*k + C

    @staticmethod
    def FVU(params,k,h,d,v):
        fi = Surface.F(params,k,h,d)
        SStot = sum((v-np.mean(v))**2)
        SSres = sum((v-fi)**2)
        return SSres/SStot

    def get_kmin(self,H,D):
        A,B,C = self.coeff(self.fit,H,D)
        return -B/(2*A)

    def eval(self,k,h,d):
        return self.F(self.fit,k,h,d)

    def R2(self):
        return 1.0 - self.FVU(self.fit,self.data[0],self.data[1],self.data[2],self.data[3])

    def R2bar(self):
        ndata = len(self.data[2])
        nparams = len(self.fit)
        return 1.0 - ((ndata-1.0)/(ndata-nparams-1.0)*(1.0 - self.R2()))

    def do_fit(self,k,h,d,v):
        ## Minimize sqrt(L2 norm) to surface (more generalizable)
        self.data = (k,h,d,v)
        res = scipy.optimize.minimize(self.FVU,self.paramguess,args=self.data)
        self.paramguess = res.x
        self.fit = res.x
        return res.x

def smooth(x,y,new_x=None):
    '''
    Smooths a function, optionally interpolates to new
    x values.
    '''
    if new_x is None:
        new_x = x
    interp0 = interp.interp1d(x,y,fill_value='extrapolate')
    x1 = np.linspace(new_x.min(),new_x.max(),600)
    y1 = interp0(x1)
    x2 = x1[15:-14]
    y2 = np.convolve(y1,np.ones(30)/30.0,mode='valid')
    interp1 = interp.interp1d(x2,y2, fill_value='extrapolate')
    return interp1(new_x)

def loadmodel(m, at_displacements):
    '''
    Loads time and interface velocity sequences from a model,
    returns values at specified interface displacement checkpoints.
    '''
    at_displacements = np.array(at_displacements)
    depth = m.get_property_sequence('position[1]', lambda y: 1.0 - y.min())

    time  = m.get_time_sequence()
    vel   = m.get_property_sequence('velocity[1]', lambda vy: np.sqrt((vy**2)).max())

    tinterp = interp.interp1d(depth-0.15, time, fill_value='extrapolate')
    T = tinterp(at_displacements)
    T[np.where(at_displacements > depth.ptp())] = np.nan

    V = smooth(depth-0.15, vel, at_displacements)
    V[np.where(at_displacements > depth.max()-0.15)] = np.nan

    return (m.dip, m.eta, m.k, T, V)

def fig3_subplotrow(K,ETA,H,D,V,disp,ax0,ax1,ax2,surf,show_xticks=True,show_titles=True):
    X,Y = np.mgrid[np.log10(0.2):np.log10(5):500j,0.0:0.9:501j]
    mnt,mxt = .05,4.8
    ext = (X.min(),X.max(),1-Y.max(),1-Y.min())
    yy = np.linspace(Y.min(),Y.max(),100)

    w = np.where(np.isfinite(V))
    if w[0].size == 0: return
    surf.do_fit(np.log10(K[w]), H[w], D[w], np.log10(1.0/V[w]))
    print 'R^2, Adjusted R^2: ', surf.R2(), surf.R2bar()

    RATE_H = 1.0 / (10**surf.eval(X,Y,0*Y))
    RATE_D = 1.0 / (10**surf.eval(X,0*Y,Y))
    RATIO = (RATE_H - RATE_D) / (RATE_H + RATE_D)

    im0 = ax0.imshow(RATE_H.T,  extent=ext, origin='upper', cmap=cmap1d, vmin=mnt, vmax=mxt,norm=LogNorm())
    ax0.plot(surf.get_kmin(yy,0*yy),1.0 - yy,'k:',linewidth=2)
    ax0.scatter(np.log10(K[w]),1.0-H[w],s=0.5,c='k')
    ax0.set_yticks(np.arange(0.1,1.1,0.1))
    ax0.set_yticklabels(['','0.2','','0.4','','0.6','','0.8','',''])
    ax0.text(np.log10(0.3),0.95,'d = %.02f'%disp, \
            fontsize=14,fontweight='bold', \
            verticalalignment='top',horizontalalignment='left')

    im1 = ax1.imshow(RATE_D.T,  extent=ext, origin='upper', cmap=cmap1d, vmin=mnt, vmax=mxt,norm=LogNorm())
    ax1.plot(surf.get_kmin(0*yy,yy),1.0 - yy,'k--',linewidth=2)
    ax1.scatter(np.log10(K[w]),1.0-D[w],s=0.5,c='k')
    ax1.set_yticks(np.arange(0.1,1.1,0.1))
    ax1.tick_params(labelleft=False)

    im2 = ax2.imshow(RATIO.T*0,cmap=cmap2d, vmin=-0.4, vmax=0.4)
    ax2.contourf(X,1.0-Y,RATIO,levels=np.linspace(-0.4,0.4,16)[1:-1], cmap=cmap2, vmin=-0.4, vmax=0.4)
    ax2.contour(X,1.0-Y,RATIO,colors='k',linewidths=0.7,levels=np.linspace(-0.4,0.4,16)[1:-1])
    ax2.scatter(np.log10(K[w]),ETA[w],s=0.5,c='k')
    ax2.plot(surf.get_kmin(yy,0*yy),1.0 - yy,'k:',linewidth=2)
    ax2.plot(surf.get_kmin(0*yy,yy),1.0 - yy,'k--',linewidth=2)
    ax2.set_yticks(np.arange(0.1,1.1,0.1))
    ax2.tick_params(labelleft=False)

    if show_titles:
        ax0.set_title('Horizontal (H)')
        ax1.set_title('Dipping (D)')
        ax2.set_title('(H-D)/(H+D)')

    for ax in (ax0,ax1,ax2):
        ax.tick_params(top=False, right=False)
        ax.set_xlim((np.log10(0.25),np.log10(5)))
        ax.set_ylim((0.1,1.0))
        ax.set_aspect(0.77255, adjustable='box')
        ticklocs = np.log10(np.arange(np.ceil(10**X.min()*10)/10.0,10**X.max(),0.1))
        if show_xticks:
            ax.set_xticks(ticklocs)
            ax.set_xticklabels(['%g'%(10**x) if np.log2(np.round((10**x),1))%1.0 == 0.0 else '' for x in ticklocs])
        else:
            ax.set_xticks([x for x in ticklocs if np.log2(np.round((10**x),1))%1.0 == 0.0])
            ax.tick_params(labelbottom=False)
    return im0, im1, im2
        
def mkfig3(ensemble):
    '''
    Plots maps of instability growth rate at different displacement checkpoints.
    '''
    surf = Surface()
    checkpoints = np.array([0.0149, 0.08, 0.15])
    DATA = [loadmodel(m, checkpoints) for m in ensemble.get_models()]
    DIP,ETA,K,TIME,VEL = [np.array(a) for a in zip(*DATA)]

    H = (1.0 - ETA)*(DIP ==  0.0)
    D = (1.0 - ETA)*(DIP == 45.0)


    try:
        pl.style.use(['classic', 'paper'])
    except: pass
    fig = pl.figure(figsize=(7.125,4.0))
    for idx, d in enumerate(checkpoints):
        ax0 = fig.add_subplot(3,3,3*idx + 1)
        ax1 = fig.add_subplot(3,3,3*idx + 2)
        ax2 = fig.add_subplot(3,3,3*idx + 3)
        im0, im1, im2 = fig3_subplotrow(K,ETA,H,D,VEL.T[idx], d, ax0, ax1, ax2, surf,\
                show_xticks=(idx==3), show_titles=(idx==0))

    pl.tight_layout(w_pad=0.5, h_pad=0.5)

    fig.subplots_adjust(left=0.09,bottom=0.09,right=0.91)

    cbar_ax1 = fig.add_axes([0.915, 0.53, 0.025, 0.4])
    cb1 = fig.colorbar(im0, cax=cbar_ax1)
    #cb1.set_label('     Growth Rate  (Disp./Time)',labelpad=0)
    cb1.set_label('Growth Rate',labelpad=0)

    cbar_ax2 = fig.add_axes([0.915, 0.105, 0.025, 0.4])
    cb2 = fig.colorbar(im2, ticks=[-0.4,0,0.4], cax=cbar_ax2)
    cb2.set_label('Relative Growth Rate',labelpad=0)

    fig.text(0.02, 0.5, 'Anisotropy Magnitude $(\eta_S/\eta_N)$', va='center', rotation='vertical', fontsize=12)
    fig.text(0.5, 0.02, 'Wave Number (k)', ha='center', fontsize=12)

    fig.savefig('fig3.eps')

def mkfig4(ensemble):
    '''
    Plots maximum interface displacement, and dominant wavenumber
    as functions of time.
    '''
    surf = Surface()
    sample_disp = np.linspace(0.015,0.135,200)
    DATA = [loadmodel(m, sample_disp) for m in ensemble.get_models()]
    DIP,ETA,K,TIME,VEL = [np.array(a) for a in zip(*DATA)]
    H = (1.0 - ETA)*(DIP== 0.0)
    D = (1.0 - ETA)*(DIP==45.0)

    hkmaxs, dkmaxs, ikmaxs = [], [], []
    hts, dts, its = [], [], []
    for idx in range(VEL.T.shape[0]):
        Tx = TIME.T[idx]
        Tw = np.where(np.isfinite(Tx))
        Vx = VEL.T[idx]
        Vw = np.where(np.isfinite(Vx))
        if Tw[0].size > 0:
            surf.do_fit(np.log10(K[Vw]), H[Vw], D[Vw], np.log10(1.0/Vx[Vw]))
            ikmax = 10**surf.get_kmin(0.0,0.0)
            hkmax = 10**surf.get_kmin(0.9,0.0)
            dkmax = 10**surf.get_kmin(0.0,0.9)

            surf.do_fit(np.log10(K[Tw]), H[Tw], D[Tw], Tx[Tw])
            it = surf.eval(np.log10(ikmax),0.0,0.0)
            ht = surf.eval(np.log10(hkmax),0.9,0.0)
            dt = surf.eval(np.log10(dkmax),0.0,0.9)
        else:
            hkmax = np.nan
            dkmax = np.nan
            ikmax = np.nan
            ht = np.nan
            dt = np.nan
            it = np.nan
        hkmaxs.append(hkmax)
        dkmaxs.append(dkmax)
        ikmaxs.append(ikmax)
        hts.append(ht)
        dts.append(dt)
        its.append(it)

    try:
        pl.style.use('paper')
    except: pass
    fig, ax0 = pl.subplots()
    ax1 = ax0.twinx()

    l1 = ax0.plot(its, sample_disp, 'k-')
    l2 = ax0.plot(hts, sample_disp, 'k:')
    l3 = ax0.plot(dts, sample_disp, 'k--')
    ax0.set_xlabel('Time')
    ax0.set_ylabel('Interface Displacement')

    ax1.plot(its,ikmaxs,'g-')
    ax1.plot(hts,hkmaxs,'g--')
    ax1.plot(dts,dkmaxs,'g:')
    ax1.set_ylabel('Dominant Wavenumber ($k_{max}$)', color='g')
    ax1.tick_params('y', colors='g')

    symb=[l[0] for l in (l1,l2,l3)]
    ax1.legend(symb,['Isotropic','Horizontal','Dipping'],loc='lower right')

    pl.tight_layout(pad=0.2)
    fig.savefig('fig4.eps')

if __name__ == '__main__':
    model_ensemble = load_ensemble()

    mkfig3(model_ensemble)

    mkfig4(model_ensemble)
