#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Generates figure (A1) for manufactured solution
convergence. Must run models in Convergence directory
before running this script:

cd Convergence
make all

'''

import pylab as pl
import numpy as np
import glob, os

##
## Subfigure A:
##

X, Y = pl.meshgrid(pl.arange(-0.5*pl.pi, 0.5 * pl.pi, .15), pl.arange(-0.5*pl.pi, 0.5 * pl.pi, .15))
U = pl.sin(2*Y)*pl.sin(2*X)
V = (pl.cos(2*Y)+1)*pl.cos(2*X)

try:
    pl.style.use(['classic', 'paper'])
except: pass
fig = pl.figure()

ax0 = fig.add_subplot(121)
ax0.quiver(Y, X, V, U, np.sqrt(U*U + V*V))
ax0.set_xlim([-pl.pi/2,pl.pi/2])
ax0.set_ylim([-pl.pi/2,pl.pi/2])

##
## Subfigure B:
##

l1normcol = -1
l2normcol = -1
with open(glob.glob('Convergence/build/*/statistics')[0],'r') as representative_stats_file:
    for i, line in enumerate(representative_stats_file.readlines()):
        if line.endswith('Error L1 Norm'):
            l1normcol = i
        if line.endswith('Error L2 Norm'):
            l2normcol = i
        if (l1normcol > -1) and (l2normcol > -1):
            break

assert (l1normcol > -1), 'No L_1 Norm statistic found'
assert (l2normcol > -1), 'No L_2 Norm statistic found'

iL1, hL1, dL1 = [], [], []
iL2, hL2, dL2 = [], [], []

ax1 = fig.add_subplot(122)
for statsfile in glob.glob('Convergence/build/*/statistics'):
    l1, l2 = pl.loadtxt(statsfile, usecols=(l1normcol, l2normcol)).T
    params = os.path.basename(os.path.dirname(statsfile)).split(',')
    refinement = int(params[0])
    dip = float(params[1])*2*pl.pi/360
    eta_s = float(params[2])

    h = 1.0/(2**refinement)
    if eta_s == 1.0:
        iL1.append((h, l1))
        iL2.append((h, l2))
    elif dip == 0:
        hL1.append((h, l1))
        hL2.append((h, l2))
    else:
        dL1.append((h, l1))
        dL2.append((h, l2))

h, iL1 = pl.array(iL1).T
ax1.plot(1.0/h, iL1, 'k--o')
h, iL2 = pl.array(iL2).T
ax1.plot(1.0/h, iL2, 'k-o')

h, hL1 = pl.array(hL1).T
ax1.plot(1.0/h, hL1, 'b--o')
h, hL2 = pl.array(hL2).T
ax1.plot(1.0/h, hL2, 'b-o')

h, dhL1 = pl.array(dL1).T
ax1.plot(1.0/h, dL1, 'r--o')
h, dL2 = pl.array(dL2).T
ax1.plot(1.0/h, dL2, 'r-o')

ax1.set_xscale('log')
ax1.set_yscale('log')
ax1.set_xlabel('$1/h_k$')
ax1.set_ylabel('Error')

black_patch = pl.matplotlib.patches.Patch(color='black', label='Isotropic')
blue_patch = pl.matplotlib.patches.Patch(color='blue',   label='Horizontal')
red_patch = pl.matplotlib.patches.Patch(color='red',     label='Dipping')
solid = pl.matplotlib.lines.Line2D([], [], color='black', marker='o', markersize=8, label='$L_2$ Misfit',linestyle='-')
dashed = pl.matplotlib.lines.Line2D([], [], color='black', marker='o', markersize=8, label='$L_1$ Misfit',linestyle='--')
ax1.legend(handles=[black_patch, blue_patch, red_patch,solid,dashed], loc='upper right')

fig.savefig('figA1.eps')
