#!/usr/bin/env python2
# -*- coding: utf-8 -*-

'''
Tools for working with model results.
'''

import os
import glob
import cPickle as pickle
import numpy as np
import warnings

class Model:
    '''
    Class to represent a single model result.
    Abstracts the time progression of particle properties within
    each model for simpler analysis.
    '''
    def __init__(self, directory):
        ## Extract info from the model directory name
        self.dir = directory
        self.name = os.path.basename(directory)

        dim,dip,eta,w = map(float, self.name.split(',')[:4])
        self.dim = dim
        self.dip = dip
        self.eta = eta
        self.k = 1/w

        statsfname = os.path.join(self.dir,'statistics')
        self.stats = open(statsfname,'r').read()

        ## Load data from model output
        self.particle_fields = set()
        self.time_sequence = []
        self.particle_sequence = []
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.load_data()

        # Convert important things to arrays where possible
        self.time_sequence = np.array(self.time_sequence)
        for timestep in self.particle_sequence:
            for field in self.particle_fields:
                if field not in timestep:
                    timestep[field] = np.array([])
                else:
                    timestep[field] = np.array(timestep[field])

    def load_data(self):
        particle_file_base = os.path.join(self.dir,'particles','particle')

        steps,times = np.loadtxt(os.path.join(self.dir,'statistics'),usecols=(0,1)).T
        for step,time in zip(steps,times):
            ### Make a dictionary of particle properties
            properties = dict()

            # Load all particle files
            particle_file_pattern = '%s-%05d.*.txt'%(particle_file_base,step)
            for particle_filename in glob.glob(particle_file_pattern):
                # Get field names from first line of file
                field_names = open(particle_filename,'r').readline().strip().split()[1:]
                # Iterate through each particle, append its properties
                # to the properties dictionary
                for particle in np.loadtxt(particle_filename):
                    try:
                        for i, field in enumerate(field_names):
                            self.particle_fields.add(field)
                            if field in properties.keys():
                                properties[field].append(particle[i])
                            else:
                                properties[field] = [particle[i]]
                    except:
                        pass

            if len(properties.keys()) > 0:
                self.time_sequence.append(time)
                self.particle_sequence.append(properties)

    def get_name(self):
        return self.name

    def get_time_sequence (self):
        return self.time_sequence

    def get_field_names(self):
        return self.particle_fields

    def get_property_sequence (self, field, process=(lambda x: x)):
        ''' Return arbitrary arrays of time vs some field,
            optionally processed by some function. '''
        if field == 'time':
            return self.get_time_sequence()
        else:
            return np.array([process(particles[field]) \
                    for particles in self.particle_sequence])

class Ensemble:
    '''
    Class to organize and query results from a set of Model objects.

    Provides useful interface for getting subsets of models based on
    combinations of criteria.
    '''
    def __init__(self, paths):
        self.models = [Model(path) for path in paths]

    def get_models(self, conditions=(lambda dim, dip, eta, k: True)):
        ''' Return models that satisfy the conditions. '''
        matches = [m for m in self.models if conditions(m.dim,m.dip,m.eta,m.k)]
        return matches

def combine_model_results (model_basename):
    '''
    Combines model results from an ensemble of models into
    a single pickle-able python object. Reduces size of model
    results from ~4GB to ~87MB.
    '''
    def ismodeldir (path):
        # Just return whether or not the given directory has Aspect model
        # results in it.
        statsfile = os.path.join(path,'statistics')
        if os.path.isfile(statsfile) and len(open(statsfile,'r').read()) > 0:
            return True
        return False

    model_paths = [path for path in glob.glob('Ensemble/build/*') if ismodeldir(path)]
    ensemble = Ensemble(model_paths)

    try:
        import backports.lzma as lzma
        with lzma.open(model_basename+'.lzma', 'wb') as pfile:
            pickle.dump(ensemble, pfile)

    except:
        import gzip
        with gzip.open(model_basename+'.gz', 'wb') as pfile:
            pickle.dump(ensemble, pfile)
    return ensemble

def load_ensemble():
    '''
    Loads an ensemble from pickled object if available, otherwise from model
    results directory.
    '''
    model_basename = 'model_ensemble_results.p'
    if os.path.isfile(model_basename+'.lzma'):
        import backports.lzma as lzma
        with lzma.open(model_basename+'.lzma','rb') as model_f:
            model_ensemble = pickle.load(model_f)
    elif os.path.isfile(model_basename+'.gz'):
        import gzip
        with gzip.open(model_basename+'.gz','rb') as model_f:
            model_ensemble = pickle.load(model_f)
    else:
        model_ensemble = combine_model_results(model_basename)
    return model_ensemble
